/* data_protocol.c - accounting info protocol
 *
 * Copyright (C) 2006-2007 Abilis
 * Copyright (C) 2006-2007 Antek srl
 *
 * Author: Elisa Manara <e@entropika.net>
 * Last update: Wed Jan 10 10:41:55 CET 2007
 *
 * This file is part of Abilis Collector for *nix.
 *
 * Abilis Collector is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * Abilis Collector is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 *                                   
 * You should have received a copy of the GNU General Public License along                        
 * with Abilis Collector; see the file COPYING.  If not, write to the Free                        
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. 
 */  

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "define.h"
#include "data_protocol.h"

void accountingManager(struct CfgData *config, struct CfgClient *client,
    unsigned char *data, unsigned int datalen)
{
    int retval, queued;
    struct AcntQuery query;

    /* clear/initialize accounting data */
    accountingInit(client, &query);
 
    /* parse data message, fill accounting structs and query fields */
    retval = accountingParse(&query, data, datalen);
    if(retval != 0) {
        printError("[%s] Incomplete or corrupted data message from agent %s!\n",
	   client->id, client->agentHost);
	return;
    }

    /* save accounting info -> MySQL */
    accountingQueryPrepare(config, client, &query);
    if(query.incomplete) {
	printError("[%s] Query too long! Buffer size is %d\n",
	    client->id, MYSQL_MAX_QLEN);
	return;
    }

    printDebug("[%s] INSERT query to MySQL server %s\n",
	client->id, config->mysqlHost);
    printDebug("[%s] QUERY: %s\n", client->id, query.body);
    /* try mysql INSERT */
    retval=mysqlPut(config, query.body);
    /* retval == -1 => skip malformed query */
    /* retval == 2  => skip duplicate query */
    /* retval == 0  => successfull query */
    if(retval != 1) return;

    /* retval == 1 => query fails due to connection lost:
     * save it to cache file */
    queued=queueAdd(config, client, query.body);

    if(queued < 0) {
	/* cannot write queue, fatal error */
	printError("[QUEUE] Fatal error! Must exit now.\n");
	collectorTerminate(config, client);
	setShutdown(COLLECTOR_HALT, 1);
    }
}

/* clear/initialize struct - begin query body */
void accountingInit(struct CfgClient *client, struct AcntQuery *query)
{
    memset(query, 0, sizeof(*query));
    snprintf(query->body, MYSQL_MAX_QLEN, "INSERT INTO %s(",
	client->mysqlTable);

    snprintf(query->collectorID, MAX_CLIENTID_LEN, "%s", client->id);
    accountingQueryAddField(query, "collector");
    accountingQueryAddStr(query, client->id);
    accountingQueryAddField(query, "agent_host");
    accountingQueryAddStr(query, client->agentHost);
}

void accountingQueryAddBody(struct AcntQuery *q, char *str)
{
    size_t blen=strlen(q->body);
    size_t slen=strlen(str);

    if((MYSQL_MAX_QLEN - (blen+1)) < slen) {
	q->incomplete=1;
	return;
    }
    strncat(q->body, str, slen);
}

void accountingQueryAddValue(struct AcntQuery *q, char *str)
{
    size_t vlen=strlen(q->values);
    size_t slen=strlen(str);

    if((MYSQL_MAX_VLEN - (vlen+1)) < slen) {
	q->incomplete=1;
	return;
    }
    strncat(q->values, str, slen);
}

void accountingQueryAddField(struct AcntQuery *q, char *str)
{
    accountingQueryAddBody(q, str);
    accountingQueryAddBody(q, ",");
}

/* quote and add a string */
void accountingQueryAddStr(struct AcntQuery *q, char *str)
{
    size_t slen=strlen(str);
    /* buffer for string escaping, at least length*2+1 bytes long. 
     * In the worst case, each character may need to be encoded 
     * as using two bytes, plus the terminating null byte
     */
    char *e = malloc((slen*2)+1);

    if (!e) {
	q->incomplete=1;
	return;
    }

    mysql_escape_string(e, str, slen);

    accountingQueryAddValue(q, "'");
    accountingQueryAddValue(q, e);
    accountingQueryAddValue(q, "',");
    free(e);
}

#define INT_BUF 32
void accountingQueryAddInt(struct AcntQuery *q, int i)
{
    char str[INT_BUF];

    snprintf(str, INT_BUF, "%d", i);
    accountingQueryAddValue(q, str);
    accountingQueryAddValue(q, ",");
}

/* prepare INSERT query for MySQL, using collected info */
void accountingQueryPrepare(struct CfgData *config, struct CfgClient *client,
    struct AcntQuery *query)
{
    /* extended callID types */
    char extConnId[EXT_ID_LEN];
    char extPInCallId[EXT_ID_LEN];
    char extPOutCallId[EXT_ID_LEN];
    /* unique key */
    char unique[UNIQUE_KEY_LEN];

    /* build extended callIDs */
    snprintf(extConnId, EXT_ID_LEN, "%04i%02i%02i%02i%02i%02i%04i",
	query->endDT.year, query->endDT.month, query->endDT.day,
	query->endDT.hour, query->endDT.min, query->endDT.sec,
	query->callID);
    if(query->inParentCallID != 0) {
	snprintf(extPInCallId, EXT_ID_LEN, "%04i%02i%02i%02i%02i%02i%04i",
	    query->startDT.year, query->startDT.month, query->startDT.day,
	    query->startDT.hour, query->startDT.min, query->startDT.sec,
	    query->inParentCallID);
    } else {
	extPInCallId[0] = '\0';
    }

    if(query->outParentCallID != 0) {
	snprintf(extPOutCallId, EXT_ID_LEN, "%04i%02i%02i%02i%02i%02i%04i",
	    query->startDT.year, query->startDT.month, query->startDT.day,
	    query->startDT.hour, query->startDT.min, query->startDT.sec,
	    query->outParentCallID);
    } else {
	extPOutCallId[0] = '\0';
    }

    accountingQueryAddField(query, "ext_connid");
    accountingQueryAddStr(query, extConnId);
    accountingQueryAddField(query, "ext_in_parent_callid");
    accountingQueryAddStr(query, extPInCallId);
    accountingQueryAddField(query, "ext_out_parent_callid");
    accountingQueryAddStr(query, extPOutCallId);

    /* build unique key from:
     * client ID
     * current time
     * ext connection ID */
    snprintf(unique, UNIQUE_KEY_LEN, "%s%d%s",
	query->collectorID, (int) time(NULL), extConnId);
    
    accountingQueryAddField(query, "unique_check");
    accountingQueryAddStr(query, unique);

    accountingQueryAddBody(query, "insert_time) VALUES(");
    accountingQueryAddBody(query, query->values);
    accountingQueryAddBody(query, "NOW())");
}

int accountingParse(struct AcntQuery *query, unsigned char *data, unsigned int datalen)
{
    unsigned char *buf = data;
    unsigned int bleft;
    int flen;
    unsigned int fname;
    char cluster[CLUSTER_NAME_LEN];
    char datetime[DATETIME_LEN];
    char username[USERNAME_LEN];
    char gmt_diff[48];
    char disc_cause_raw[16];
    
    /* yet unread bytes */
    bleft = datalen;

    /* skip packet headers */
    bleft = parseForward(&buf, bleft, PACKET_ACNT_POSITION);
    /* unexpected end of buffer? :-S */
    if(bleft == 0) return -1;

    for (; buf < (unsigned char*) (data+datalen); buf++,bleft--) {
	/* unexpected end of buffer? */
	if(bleft == 0) return -1;

	fname = *buf;

	bleft = parseForward(&buf, bleft, 1);
	/* unexpected end of buffer? */
	if(bleft == 0) return -1;
	flen = *buf;
	bleft = parseForward(&buf, bleft, 1);
	/* unexpected end of buffer? */
	if(bleft == 0) return -1;

        printDebug("ACNT FIELD: %02x - LEN: %i\n", fname, flen);
	
	switch(fname) {
	/* 0x01 originatin port */
	case ACNT_IE_PORT_IN:
	    accountingQueryAddField(query, "orig_port");
	    accountingQueryAddInt(query, extractWordField(buf));
	    printDebug("# Originating port: %i\n", extractWordField(buf));
	    break;
	/* 0x02 incoming called number */
	case ACNT_IE_CDI:
	    memcpy(&query->inCalledNum, buf, flen);
	    accountingQueryAddField(query, 
		"in_called_num_type,in_called_num_plan,in_called_num");
	    accountingQueryAddInt(query, query->inCalledNum.type);
	    accountingQueryAddInt(query, query->inCalledNum.plan);
	    accountingQueryAddStr(query, query->inCalledNum.number);
	    printDebug("# Incoming called number: %s\n", query->inCalledNum.number);
	    break;
	/* 0x03 incoming called subaddress */
	case ACNT_IE_SDI:
	    memcpy(&query->inCalledSA, buf, flen);
	    accountingQueryAddField(query, 
		"in_called_subaddr_type,in_called_subaddr_ind,in_called_subaddr");
	    accountingQueryAddInt(query, query->inCalledSA.type);
	    accountingQueryAddInt(query, query->inCalledSA.indicator);
	    accountingQueryAddStr(query, query->inCalledSA.address);
	    printDebug("# Incoming called subaddress: %s\n", query->inCalledSA.address);
	    break;
	/* 0x04 incoming calling number */
	case ACNT_IE_CGI:
	    memcpy(&query->inCallingNum, buf, flen);
	    accountingQueryAddField(query,
		"in_calling_num_type,in_calling_num_plan");

	    accountingQueryAddInt(query, query->inCallingNum.type);
	    accountingQueryAddInt(query, query->inCallingNum.plan);

	    /* presentation/screening could be absent; in this case, 
	     * we recompose the number field 
	     */
	    if(query->inCallingNum.ext == 1) {
		memcpy( query->inCallingNum.number, &buf[ 1 ], flen - 1 );
	    } else {
		accountingQueryAddField(query, "in_calling_num_pres, in_calling_num_screen");
		accountingQueryAddInt(query, query->inCallingNum.presentation);
		accountingQueryAddInt(query, query->inCallingNum.screening);
	    }

	    accountingQueryAddField(query, "in_calling_num");
	    accountingQueryAddStr(query, query->inCallingNum.number);

	    printDebug("# Incoming calling number: %s\n", query->inCallingNum.number);
	    break;
	/* 0x05 incoming calling subaddress */
	case ACNT_IE_SGI:
	    memcpy(&query->inCallingSA, buf, flen);
	    accountingQueryAddField(query, 
		"in_calling_subaddr_type,in_calling_subaddr_ind,in_calling_subaddr");
	    accountingQueryAddInt(query, query->inCallingSA.type);
	    accountingQueryAddInt(query, query->inCallingSA.indicator);
	    accountingQueryAddStr(query, query->inCallingSA.address);
	    printDebug("# Incoming calling subaddress: %s\n", query->inCallingSA.address);
	    break;
	/* 0x06 originating cluster name */
	case ACNT_IE_EBIB_IN:
	    extractStringField(buf, flen, cluster);
	    accountingQueryAddField(query, "orig_cluster_name");
	    accountingQueryAddStr(query, cluster);
	    printDebug("# Originating cluster name: %s\n", cluster);
	    break;
	/* 0x07 originating side */
	case ACNT_IE_SIDE_IN:
	    accountingQueryAddField(query, "orig_side");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Originating side: %i\n", extractByteField(buf));
	    break;
	/* 0x08 originating port type */
	case ACNT_IE_TYPE_IN:
	    accountingQueryAddField(query, "orig_port_type");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Originating port type: %i\n", extractByteField(buf));
	    break;
	/* 0x09 incoming parent call identifier */
	case ACNT_IE_PARENT_ID_IN:
	    query->inParentCallID=extractDwordField(buf);
	    accountingQueryAddField(query, "in_parent_callid");
	    accountingQueryAddInt(query, query->inParentCallID);
	    printDebug("# Incoming parent call identifier: %i\n", 
		query->inParentCallID);
	    break;
	/* 0x0A username in */
	case ACNT_IE_USER_IN:
	    extractStringField(buf, flen, username);
	    accountingQueryAddField(query, "user_in");
	    accountingQueryAddStr(query, username);
	    printDebug(" # Username in: %s\n", username);
	    break;
	/* 0x0B redirecting number incoming */
	case ACNT_IE_RGI:
      memcpy(&query->redNumIn, buf, flen);
      accountingQueryAddField(query,
          "red_num_in_type,red_num_in_plan");

      accountingQueryAddInt(query, query->redNumIn.type);
      accountingQueryAddInt(query, query->redNumIn.plan);

      /* presentation/screening could be absent; in this case, 
       * we recompose the number field 
       */
      if(query->redNumIn.ext == 1) {
        memcpy( query->redNumIn.number, &buf[ 1 ], flen - 1 );
      } else {
        accountingQueryAddField(query, "red_num_in_pres, red_num_in_screen");
        accountingQueryAddInt(query, query->redNumIn.presentation);
        accountingQueryAddInt(query, query->redNumIn.screening);
      }

      accountingQueryAddField(query, "red_num_in");
      accountingQueryAddStr(query, query->redNumIn.number);

      printDebug("# Redirecting number incoming: %s\n", query->redNumIn.number);
      break;
	/* 0x11 destination port */
	case ACNT_IE_PORT_OUT:
	    accountingQueryAddField(query, "dest_port");
	    accountingQueryAddInt(query, extractWordField(buf));
	    printDebug("# Destination port: %i\n", extractWordField(buf));
	    break;
	/* 0x12 outgoing called number */
	case ACNT_IE_CDO:
	    memcpy(&query->outCalledNum, buf, flen);
	    accountingQueryAddField(query, 
		"out_called_num_type,out_called_num_plan,out_called_num");
	    accountingQueryAddInt(query, query->outCalledNum.type);
	    accountingQueryAddInt(query, query->outCalledNum.plan);
	    accountingQueryAddStr(query, query->outCalledNum.number);
	    printDebug("# Outgoing called number: %s\n", query->outCalledNum.number);
	    break;
	/* 0x13 outgoing called subaddress */
	case ACNT_IE_SDO:
	    memcpy(&query->outCalledSA, buf, flen);
	    accountingQueryAddField(query, 
		"out_called_subaddr_type,out_called_subaddr_ind,out_called_subaddr");
	    accountingQueryAddInt(query, query->outCalledSA.type);
	    accountingQueryAddInt(query, query->outCalledSA.indicator);
	    accountingQueryAddStr(query, query->outCalledSA.address);
	    printDebug("# Outgoing called subaddress: %s\n", query->outCalledSA.address);
	    break;
	/* 0x14 outgoing calling number */
	case ACNT_IE_CGO:
	    memcpy(&query->outCallingNum, buf, flen);
	    accountingQueryAddField(query,
		"out_calling_num_type,out_calling_num_plan");

	    accountingQueryAddInt(query, query->outCallingNum.type);
	    accountingQueryAddInt(query, query->outCallingNum.plan);

	    /* presentation/screening could be absent; in this case, 
	     * we recompose the number field 
	     */
	    if(query->outCallingNum.ext == 1) {
		memcpy( query->outCallingNum.number, &buf[ 1 ], flen - 1 );
	    } else {
		accountingQueryAddField(query, "out_calling_num_pres, out_calling_num_screen");
		accountingQueryAddInt(query, query->outCallingNum.presentation);
		accountingQueryAddInt(query, query->outCallingNum.screening);
	    }

	    accountingQueryAddField(query, "out_calling_num");
	    accountingQueryAddStr(query, query->outCallingNum.number);

	    printDebug("# Outgoing calling number: %s\n", query->outCallingNum.number);
	    break;
	/* 0x15 outgoing calling subaddress */
	case ACNT_IE_SGO:
	    memcpy(&query->outCallingSA, buf, flen);
	    accountingQueryAddField(query, 
		"out_calling_subaddr_type,out_calling_subaddr_ind,out_calling_subaddr");
	    accountingQueryAddInt(query, query->outCallingSA.type);
	    accountingQueryAddInt(query, query->outCallingSA.indicator);
	    accountingQueryAddStr(query, query->outCallingSA.address);
	    printDebug("# Outgoing calling subaddress: %s\n", query->outCallingSA.address);
	    break;
	/* 0x16 destination cluster name */
	case ACNT_IE_EBIB_OUT:
	    extractStringField(buf, flen, cluster);
	    accountingQueryAddField(query, "dest_cluster_name");
	    accountingQueryAddStr(query, cluster);
	    printDebug("# Destination cluster name: %s\n", cluster);
	    break;
	/* 0x17 destination side */
	case ACNT_IE_SIDE_OUT:
	    accountingQueryAddField(query, "dest_side");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Destination side: %i\n", extractByteField(buf));
	    break;
	/* 0x18 destination port type */
	case ACNT_IE_TYPE_OUT:
	    accountingQueryAddField(query, "dest_port_type");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Destination port type: %i\n", extractByteField(buf));
	    break;
	/* 0x19 outgoing parent call identifier */
	case ACNT_IE_PARENT_ID_OUT:
	    query->outParentCallID=extractDwordField(buf);
	    accountingQueryAddField(query, "out_parent_callid");
	    accountingQueryAddInt(query, query->outParentCallID);
	    printDebug("# Outgoing parent call identifier: %i\n",
		query->outParentCallID);
	    break;
	/* 0x1A username out */
	case ACNT_IE_USER_OUT:
	    extractStringField(buf, flen, username);
	    accountingQueryAddField(query, "user_out");
	    accountingQueryAddStr(query, username);
	    printDebug(" # Username out: %s\n", username);
	    break;
	/* 0x1B redirecting number outgoing */
	case ACNT_IE_RGO:
      memcpy(&query->redNumOut, buf, flen);
      accountingQueryAddField(query,
          "red_num_out_type,red_num_out_plan");

      accountingQueryAddInt(query, query->redNumOut.type);
      accountingQueryAddInt(query, query->redNumOut.plan);

      /* presentation/screening could be absent; in this case, 
       * we recompose the number field 
       */
      if(query->redNumOut.ext == 1) {
        memcpy( query->redNumOut.number, &buf[ 1 ], flen - 1 );
      } else {
        accountingQueryAddField(query, "red_num_out_pres, red_num_out_screen");
        accountingQueryAddInt(query, query->redNumOut.presentation);
        accountingQueryAddInt(query, query->redNumOut.screening);
      }

      accountingQueryAddField(query, "red_num_out");
      accountingQueryAddStr(query, query->redNumOut.number);

      printDebug("# Redirecting number outgoing: %s\n", query->redNumOut.number);
      break;
	/* 0x21 connection type */
	case ACNT_IE_CONNECTION_TYPE:
	    accountingQueryAddField(query, "conn_type");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Connection type: %i\n", extractByteField(buf));
	    break;
	/* 0x22 disconnection cause */
	case ACNT_IE_CAUSE:
	    memcpy(&query->dc, buf, flen);
	    /*
	    accountingQueryAddField(query, 
		"disc_coding,disc_location,disc_recom,disc_cause,disc_diagnostic");
	    */

	    //printDump(buf, flen);
	    accountingQueryAddField(query, 
		"disc_coding,disc_location,disc_cause, disc_cause_raw");
	    accountingQueryAddInt(query, query->dc.coding);
	    accountingQueryAddInt(query, query->dc.location);
	    accountingQueryAddInt(query, query->dc.cause);
	    hexString(buf, disc_cause_raw, flen);
	    accountingQueryAddStr(query, disc_cause_raw);
	    //accountingQueryAddInt(query, query->dc.diagnostic);
	    //accountingQueryAddInt(query, query->dc.recommendation & 0x0F);
	    printDebug("# Disconnection location: %i\n", query->dc.location);
	    printDebug("# Disconnection cause: %i\n", query->dc.cause);
	    break;
	/* 0x23 bearer capability */
	case ACNT_IE_BEARER:
	    memcpy(&query->bearer, buf, flen);
	    accountingQueryAddField(query, "bearer_codec");
	    accountingQueryAddInt(query, query->bearer.type);
	    accountingQueryAddField(query, "bearer_note");
	    accountingQueryAddInt(query, query->bearer.note);
	    if(query->bearer.bflag == 0) {
		accountingQueryAddField(query, "bearer_bitrate");
		accountingQueryAddInt(query, query->bearer.bitrate);
	    }
	    printDebug("# Bearer capability: %i - (%i), %i - %i\n", 
		query->bearer.type,
		query->bearer.bflag,
		query->bearer.bitrate,
		query->bearer.note);
	    break;
	/* 0x24 call start date/time */
	case ACNT_IE_INCOMING_DATE_TIME:
	    memcpy(&query->startDT, buf, flen);
	    formatDateTime(&query->startDT, datetime);
	    accountingQueryAddField(query, "call_start");
	    accountingQueryAddStr(query, datetime);

	    accountingQueryAddField(query, "call_start_gmt");
	    snprintf(gmt_diff, sizeof(gmt_diff), 
		"(call_start + INTERVAL '%s%d:%d' HOUR_MINUTE),",
		(query->startDT.sign == 1) ? "+" : "-",
		query->startDT.GMThour, query->startDT.GMTmin);
	    accountingQueryAddValue(query, gmt_diff);

	    printDebug("# Call start: %s\n", datetime);
	    break;
	/* 0x25 call end date/time */
	case ACNT_IE_END_DATE_TIME:
	    memcpy(&query->endDT, buf, flen);
	    formatDateTime(&query->endDT, datetime);
	    accountingQueryAddField(query, "call_end");
	    accountingQueryAddStr(query, datetime);

	    accountingQueryAddField(query, "call_end_gmt");
	    snprintf(gmt_diff, sizeof(gmt_diff), 
		"(call_end + INTERVAL '%s%d:%d' HOUR_MINUTE),",
		(query->endDT.sign == 1) ? "+" : "-",
		query->endDT.GMThour, query->endDT.GMTmin);
	    accountingQueryAddValue(query, gmt_diff);

	    printDebug("# Call end: %s\n", datetime);
	    break;
	/* 0x26 connection state at disconnection */
	case ACNT_IE_CONNECTION_STATE:
	    accountingQueryAddField(query, "disc_conn_state");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Connection state at disconnection: %i\n", 
		extractByteField(buf));
	    break;
	/* 0x27 call duration */
	case ACNT_IE_CALL_LENGTH:
	    accountingQueryAddField(query, "call_time");
	    accountingQueryAddInt(query, extractWordField(buf));
	    printDebug("# Call duration: %i\n", extractWordField(buf));
	    break;
	/* 0x28 disconnection direction */
	case ACNT_IE_DISCONNECT_DIRECTION:
	    accountingQueryAddField(query, "disc_direction");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Disconnection direction: %i\n", 
		extractByteField(buf));
	    break;
	/* 0x29 call disconnection date/time */
	case ACNT_IE_DISCONNECT_DATE_TIME:
	    memcpy(&query->discDT, buf, flen);
	    formatDateTime(&query->discDT, datetime);
	    accountingQueryAddField(query, "call_disc");
	    accountingQueryAddStr(query, datetime);

	    accountingQueryAddField(query, "call_disc_gmt");
	    snprintf(gmt_diff, sizeof(gmt_diff), 
		"(call_disc + INTERVAL '%s%d:%d' HOUR_MINUTE),",
		(query->discDT.sign == 1) ? "+" : "-",
		query->discDT.GMThour, query->discDT.GMTmin);
	    accountingQueryAddValue(query, gmt_diff);

	    printDebug("# Call disconnection: %s\n", datetime);
	    break;
	/* 0x2A result of the call */
	case ACNT_IE_CALL_RESULT:
	    accountingQueryAddField(query, "call_result");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Call result: %i\n", extractByteField(buf));
	    break;
	/* 0x2B call direction */
	case ACNT_IE_CALL_DIRECTION:
	    accountingQueryAddField(query, "call_direction");
	    accountingQueryAddInt(query, extractByteField(buf));
	    printDebug("# Call direction: %i\n", extractByteField(buf));
	    break;
	/* 0x2C call identifier */
	case ACNT_IE_CALL_ID:
	    query->callID=extractDwordField(buf);
	    accountingQueryAddField(query, "callid");
	    accountingQueryAddInt(query, query->callID);
	    printDebug("# Call ID: %i\n", query->callID);
	    break;
	/* 0x41 speech compression algorithm and bit rate
	 * Simple1k only - not implemented */
	case ACNT_IE_VOICE_COMPRESSION:
	    break;
	/* 0x42 fax (or Fax/Modem with AC4801F-C) 
	 * relayed during the call and speed 
	 * Simple1k only - not implemented */
	case ACNT_IE_FAX_RELAY:
	    break;
	/* 0x43 Modem relayed during the call and speed (AC481xxA-C only)
	 * Simple1k only - not implemented */
	case ACNT_IE_MODEM_RELAY:
	    break;
	/* 0x44 audio codec properties */
	case ACNT_IE_NEW_VOICE_COMPRESSION:
	    memcpy(&query->aCodec, buf, flen);
	    accountingQueryAddField(query, 
		"audio_law,audio_ss,audio_codec,audio_bitrate,audio_bandwidth");
	    accountingQueryAddInt(query, query->aCodec.aflag);
	    accountingQueryAddStr(query, (query->aCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->aCodec.codec);
	    accountingQueryAddInt(query, query->aCodec.bitrate);
	    accountingQueryAddInt(query, query->aCodec.bandwidth);
	    printDebug("# Audio law: %i ss: %i codec type: %i\n", 
		query->aCodec.aflag, query->aCodec.bflag, query->aCodec.codec);
	    break;
	/* 0x45 fax codec properties */
	case ACNT_IE_NEW_FAX_RELAY:
	    memcpy(&query->fCodec, buf, flen);
	    accountingQueryAddField(query, 
		"fax_relay,fax_bypass,fax_codec,fax_bitrate,fax_bandwidth");
	    accountingQueryAddStr(query, (query->fCodec.aflag == 0) ? "n" : "y");
	    accountingQueryAddStr(query, (query->fCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->fCodec.codec);
	    accountingQueryAddInt(query, query->fCodec.bitrate);
	    accountingQueryAddInt(query, query->fCodec.bandwidth);
	    printDebug("# Fax codec type: %i\n", query->fCodec.codec);
	    break;
	/* 0x46 data codec properties */
	case ACNT_IE_NEW_MODEM_RELAY:
	    memcpy(&query->dCodec, buf, flen);
	    accountingQueryAddField(query, 
		"data_relay,data_bypass,data_codec,data_bitrate,data_bandwidth");
	    accountingQueryAddStr(query, (query->dCodec.aflag == 0) ? "n" : "y");
	    accountingQueryAddStr(query, (query->dCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->dCodec.codec);
	    accountingQueryAddInt(query, query->dCodec.bitrate);
	    accountingQueryAddInt(query, query->dCodec.bandwidth);
	    printDebug("# Data codec type: %i\n", query->dCodec.codec);
	    break;
	/* 0x47 reserved bandwidth */
	case ACNT_IE_RESERVED_BANDWIDTH:
	    accountingQueryAddField(query, "reserved_bandwidth");
	    accountingQueryAddInt(query, extractDwordField(buf));
	    printDebug("# Reserved bandwidth: %i\n", extractDwordField(buf));
	    break;
	/* 0x48 local session informations */
	case ACNT_IE_LOCAL_AC_SES_INFO:
	    memcpy(&query->session_info, buf, flen);
	    accountingQueryAddField(query,
		"local_voice_underrun,local_voice_overrun,local_fax_underrun,local_fax_overrun,"
		"local_voice_def_jitter,local_voice_max_jitter,local_voice_top_jitter,"
		"local_voice_avg_jitter,"
		"local_fax_def_jitter,local_fax_max_jitter,local_fax_top_jitter,"
		"local_fax_avg_jitter,local_fax_tx_pages");
	    accountingQueryAddInt(query, query->session_info.dwVoiceUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwVoiceOverrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxOverrun);
	    accountingQueryAddInt(query, query->session_info.wVoiceDefJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceTopJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxDefJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTopJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTxPages);
	    printDebug("# Local session info: vUO: %i/%i, fUO: %i/%i, vDMTA: %i/%i/%i/%i,"
		" fDMTAP: %i/%i/%i/%i/%i\n",
		query->session_info.dwVoiceUnderrun, query->session_info.dwVoiceOverrun,
		query->session_info.dwFaxUnderrun, query->session_info.dwFaxOverrun,
		query->session_info.wVoiceDefJitter, query->session_info.wVoiceMaxJitter,
		query->session_info.wVoiceTopJitter, query->session_info.wVoiceAvgJitter,
		query->session_info.wFaxDefJitter, query->session_info.wFaxMaxJitter,
		query->session_info.wFaxTopJitter, query->session_info.wFaxAvgJitter,
		query->session_info.wFaxTxPages);
	    break;
	/* 0x49 remote session informations */
	case ACNT_IE_REMOTE_AC_SES_INFO:
	    memcpy(&query->session_info, buf, flen);
	    accountingQueryAddField(query,
		"remote_voice_underrun,remote_voice_overrun,remote_fax_underrun,remote_fax_overrun,"
		"remote_voice_def_jitter,remote_voice_max_jitter,remote_voice_top_jitter,"
		"remote_voice_avg_jitter,"
		"remote_fax_def_jitter,remote_fax_max_jitter,remote_fax_top_jitter,"
		"remote_fax_avg_jitter,remote_fax_tx_pages");
	    accountingQueryAddInt(query, query->session_info.dwVoiceUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwVoiceOverrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxOverrun);
	    accountingQueryAddInt(query, query->session_info.wVoiceDefJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceTopJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxDefJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTopJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTxPages);
	    printDebug("# Remote session info: vUO: %i/%i, fUO: %i/%i, vDMTA: %i/%i/%i/%i,"
		" fDMTAP: %i/%i/%i/%i/%i\n",
		query->session_info.dwVoiceUnderrun, query->session_info.dwVoiceOverrun,
		query->session_info.dwFaxUnderrun, query->session_info.dwFaxOverrun,
		query->session_info.wVoiceDefJitter, query->session_info.wVoiceMaxJitter,
		query->session_info.wVoiceTopJitter, query->session_info.wVoiceAvgJitter,
		query->session_info.wFaxDefJitter, query->session_info.wFaxMaxJitter,
		query->session_info.wFaxTopJitter, query->session_info.wFaxAvgJitter,
		query->session_info.wFaxTxPages);
	    break;

	/* 0x54 TC audio codec properties */
	case ACNT_IE_TC_AUDIO_CODEC_PROPERTIES:
	    memcpy(&query->atcCodec, buf, flen);
	    accountingQueryAddField(query, 
		"tc_audio_law,tc_audio_ss,tc_audio_codec,tc_audio_bitrate,tc_audio_bandwidth");
	    accountingQueryAddInt(query, query->atcCodec.aflag);
	    accountingQueryAddStr(query, (query->atcCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->atcCodec.codec);
	    accountingQueryAddInt(query, query->atcCodec.bitrate);
	    accountingQueryAddInt(query, query->atcCodec.bandwidth);
	    printDebug("# TC Audio law: %i ss: %i codec type: %i\n", 
		query->atcCodec.aflag, query->atcCodec.bflag, query->atcCodec.codec);
	    break;
	/* 0x55 TC fax codec properties */
	case ACNT_IE_TC_FAX_CODEC_PROPERTIES:
	    memcpy(&query->ftcCodec, buf, flen);
	    accountingQueryAddField(query, 
		"tc_fax_relay,tc_fax_bypass,tc_fax_codec,tc_fax_bitrate,tc_fax_bandwidth");
	    accountingQueryAddStr(query, (query->ftcCodec.aflag == 0) ? "n" : "y");
	    accountingQueryAddStr(query, (query->ftcCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->ftcCodec.codec);
	    accountingQueryAddInt(query, query->ftcCodec.bitrate);
	    accountingQueryAddInt(query, query->ftcCodec.bandwidth);
	    printDebug("# TC Fax codec type: %i\n", query->ftcCodec.codec);
	    break;
	/* 0x56 TC data codec properties */
	case ACNT_IE_TC_DATA_CODEC_PROPERTIES:
	    memcpy(&query->dtcCodec, buf, flen);
	    accountingQueryAddField(query, 
		"tc_data_relay,tc_data_bypass,tc_data_codec,tc_data_bitrate,tc_data_bandwidth");
	    accountingQueryAddStr(query, (query->dtcCodec.aflag == 0) ? "n" : "y");
	    accountingQueryAddStr(query, (query->dtcCodec.bflag == 0) ? "n" : "y");
	    accountingQueryAddInt(query, query->dtcCodec.codec);
	    accountingQueryAddInt(query, query->dtcCodec.bitrate);
	    accountingQueryAddInt(query, query->dtcCodec.bandwidth);
	    printDebug("# TC Data codec type: %i\n", query->dtcCodec.codec);
	    break;
	/* 0x57 reserved bandwidth */
	case ACNT_IE_TC_RESERVED_BANDWIDTH:
	    accountingQueryAddField(query, "tc_reserved_bandwidth");
	    accountingQueryAddInt(query, extractDwordField(buf));
	    printDebug("# TC Reserved bandwidth: %i\n", extractDwordField(buf));
	    break;
	/* 0x58 transcoding local session informations */
	case ACNT_IE_TC_LOCAL_AC_SES_INFO:
	    memcpy(&query->session_info, buf, flen);
	    accountingQueryAddField(query,
		"tc_local_voice_underrun,tc_local_voice_overrun,tc_local_fax_underrun,"
		"tc_local_fax_overrun,"
		"tc_local_voice_def_jitter,tc_local_voice_max_jitter,tc_local_voice_top_jitter,"
		"tc_local_voice_avg_jitter,"
		"tc_local_fax_def_jitter,tc_local_fax_max_jitter,tc_local_fax_top_jitter,"
		"tc_local_fax_avg_jitter,tc_local_fax_tx_pages");
	    accountingQueryAddInt(query, query->session_info.dwVoiceUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwVoiceOverrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxOverrun);
	    accountingQueryAddInt(query, query->session_info.wVoiceDefJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceTopJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxDefJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTopJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTxPages);
	    printDebug("# Transcoding local session info: vUO: %i/%i, fUO: %i/%i, vDMTA: %i/%i/%i/%i,"
		" fDMTAP: %i/%i/%i/%i/%i\n",
		query->session_info.dwVoiceUnderrun, query->session_info.dwVoiceOverrun,
		query->session_info.dwFaxUnderrun, query->session_info.dwFaxOverrun,
		query->session_info.wVoiceDefJitter, query->session_info.wVoiceMaxJitter,
		query->session_info.wVoiceTopJitter, query->session_info.wVoiceAvgJitter,
		query->session_info.wFaxDefJitter, query->session_info.wFaxMaxJitter,
		query->session_info.wFaxTopJitter, query->session_info.wFaxAvgJitter,
		query->session_info.wFaxTxPages);
	    break;
	/* 0x59 transcoding remote session informations */
	case ACNT_IE_TC_REMOTE_AC_SES_INFO:
	    memcpy(&query->session_info, buf, flen);
	    accountingQueryAddField(query,
		"tc_remote_voice_underrun,tc_remote_voice_overrun,tc_remote_fax_underrun,"
		"tc_remote_fax_overrun,"
		"tc_remote_voice_def_jitter,tc_remote_voice_max_jitter,tc_remote_voice_top_jitter,"
		"tc_remote_voice_avg_jitter,"
		"tc_remote_fax_def_jitter,tc_remote_fax_max_jitter,tc_remote_fax_top_jitter,"
		"tc_remote_fax_avg_jitter,tc_remote_fax_tx_pages");
	    accountingQueryAddInt(query, query->session_info.dwVoiceUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwVoiceOverrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxUnderrun);
	    accountingQueryAddInt(query, query->session_info.dwFaxOverrun);
	    accountingQueryAddInt(query, query->session_info.wVoiceDefJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceTopJitter);
	    accountingQueryAddInt(query, query->session_info.wVoiceAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxDefJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxMaxJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTopJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxAvgJitter);
	    accountingQueryAddInt(query, query->session_info.wFaxTxPages);
	    printDebug("# Transcoding remote session info: vUO: %i/%i, fUO: %i/%i, vDMTA: %i/%i/%i/%i,"
		" fDMTAP: %i/%i/%i/%i/%i\n",
		query->session_info.dwVoiceUnderrun, query->session_info.dwVoiceOverrun,
		query->session_info.dwFaxUnderrun, query->session_info.dwFaxOverrun,
		query->session_info.wVoiceDefJitter, query->session_info.wVoiceMaxJitter,
		query->session_info.wVoiceTopJitter, query->session_info.wVoiceAvgJitter,
		query->session_info.wFaxDefJitter, query->session_info.wFaxMaxJitter,
		query->session_info.wFaxTopJitter, query->session_info.wFaxAvgJitter,
		query->session_info.wFaxTxPages);
	    break;

	/* 0x7F lost records */
	case ACNT_IE_LOST_RECORDS:
	    accountingQueryAddField(query, "lost_records");
	    accountingQueryAddInt(query, extractWordField(buf));
	    printDebug("# Lost records: %i\n", extractWordField(buf));
	    break;
	default:
	    printError("Unknown accounting element: %i, length %i\n",
		fname, flen);
	    break;
	}

	bleft = parseForward(&buf, bleft, (flen-1));
    }

    return bleft|query->incomplete;
}

int parseForward(unsigned char **buf, unsigned int bleft, int jump)
{
    if(bleft <= jump)
	return 0;

    *buf += jump;
    bleft -= jump;
    return bleft;
}

int formatDateTime(struct DateTime *dt, char *buf)
{
    return snprintf(buf, DATETIME_LEN, 
	"%04i/%02i/%02i %02i:%02i:%02i",
	dt->year, dt->month, dt->day,
	dt->hour, dt->min, dt->sec);
}

int extractByteField(unsigned char *field)
{
    //return field[0]&0x3F;
    return field[0];
}

int extractWordField(unsigned char *field)
{
    return field[1]<<8 | field[0];
}

int extractDwordField(unsigned char *field)
{
    return field[0] | field[1]<<8 | field[2]<<16 | field[3]<<24;
}

int extractStringField(unsigned char *field, int flen, char *string)
{
    return snprintf(string, flen+1, "%s", (char *) field);
}

/* 
 * vim: softtabstop=4 shiftwidth=4 
 */
